/* exec2.c -- bytecode interpreter main loop
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:39 by piumarta on emilia-2.local
*/

#define CACHE_SP	1	// 1 -> currentContext->sp cached in local register

static oop backtrace;

static void printBacktrace(void)
{
  int n= 0;
  oop cx= backtrace;
  for (;;) {
    switch (getTypeIndex(cx)) {
    case iPair: {
      oop fn= getHead(cx);
      printf("%04d ", n++);
      println(fn);
      cx= getTail(cx);
      break;
    }
    case iContext: {
      oop fn= get(cx, Context,function);
      if (notNil(fn)) {
	printf("%04d ", n++);
	if (notNil(get(fn, Expr,name)))
	  println(get(fn, Expr,name));
	else {
	  printf("(lambda ");
	  print(get(get(cx, Context,function), Expr,params));
	  fn= get(get(cx, Context,function), Expr,body);
	  while (isPair(fn)) {
	    printf(" ");
	    print(getHead(fn));
	    fn= getTail(fn);
	  }
	  printf(")\n");
	}
      }
      cx= get(cx, Context,caller);
      break;
    }
    default:
      return;
    }
  }
}

static oop apply(oop fn, oop args, oop env);

static void dumpi(oop program)
{
  oop insn= car(program);
  if (!isPair(insn)) {
    printf("ILLEGAL INSN (1) ");
    println(insn);
  } else {
    oop opcode= car(insn);
    oop operand= cdr(insn);
    if (notNil(assq(mkSymbol("branch-target"), getProperties(program))))
      printf("\t%08lx", (long)program);
    else
      printf("\t--------");
    printf(" (");
    print(opcode);
    printf(" . ");
    if (opcode == s_br || opcode == s_bt || opcode == s_bf) {
      printf("%p", operand);
    } else {
      print(operand);
    }
    printf(")");
  }
}

static void dump(char *prefix, oop program)
{
  for (;  isPair(program);  program= cdr(program)) {
    printf("%s", prefix);
    dumpi(program);
    printf("\n");
  }
  if (nil != program) {
    printf("ILLEGAL INSN (2) ");
    println(program);
  }
}

static word objectToWord(oop obj)
{
  switch (getTypeIndex(obj)) {
  case iUndefined:	return 0;
  case iLong:		return (word)getLong(obj);
  case iPointer:	return (word)get(obj, Pointer,address);
  case iString:		return (word)get(obj, String,elements);
  case iWordArray:	return (word)get(obj, WordArray,elements);
  case iSymbol:		return (word)get(get(obj, Symbol,string), String,elements);
  case iSubr:		return (word)get(obj, Subr,exec);
  default:
    fprintf(stderr, "\ncannot convert ");
    fprint(obj, stderr);
    fprintf(stderr, " to word\n");
    break;
  }
  return 0;
}

static void objectToPrimitive(oop obj, long *args, int *index, oop type)
{
  long arg;
  switch (getTypeIndex(obj)) {
  case iUndefined:	arg= 0;					break;
  case iLong:		arg= getLong(obj);			break;
  case iPointer:	arg= (long)get(obj, Pointer,address);	break;
  case iFloat: {
    double d= getFloat(obj);
    memcpy(args + *index, &d, sizeof(d));
    *index += sizeof(d) / sizeof(long);
    return;
  }
  case iString:		arg= (long)get(obj, String,elements);				break;
  case iSymbol:		arg= (long)get(get(obj, Symbol,string), String,elements);	break;
  case iWordArray:	arg= (long)get(obj, WordArray,elements);			break;
  case iSubr:		arg= (long)get(obj, Subr,exec);					break;
  default:
    if (type != s_oop) {
      fprintf(stderr, "\npassing ");
      fprint(obj, stderr);
      fprintf(stderr, " to external function\n");
    }
    arg= (long)obj;
    break;
  }
  args[(*index)++]= arg;
}

#define STACK_DIRECTION		-1

static oop *growStack(oop context)
{
  oop    *base    = get(context, Context,stackBase);
  oop    *limit   = get(context, Context,stackLimit);
  size_t  size    = (char *)limit - (char *)base;
  char   *newBase = (char *)GC_malloc(size * 2);
  oop    *sp      = (oop *)(newBase + size);
  memcpy(sp, base, size);
  set(context, Context,stackBase,  (oop *)(newBase             ));
  set(context, Context,stackLimit, (oop *)(newBase + (size * 2)));
  return sp;
}

static inline void _push(oop context, oop obj)
{
  oop *sptr= get(context, Context,sp);
  if (sptr == get(context, Context,stackBase)) sptr= growStack(context);
  *--sptr= obj;
  set(context, Context,sp, sptr);
}

#define push(OBJ)	_push(context, OBJ)
#define nos(N)		(get(context, Context,sp)[N])
#define tos()		nos(0)
#define drop(N)		(set(context, Context,sp, get(context, Context,sp) + (N)))
#define pop()		drop(1)
#define stackSize()	(get(context, Context,stackLimit) - get(context, Context,sp))

static oop makeArgs(oop *ap, int nArgs)
{
  oop args= nil;
  GC_PROTECT(args);
  ap -= nArgs * STACK_DIRECTION;
  while (nArgs--) {
    ap += STACK_DIRECTION;
    args= cons(*ap, args);
  }
  GC_UNPROTECT(args);
  return args;
}

static void initStack(oop context, oop arguments)
{
  if (isPair(arguments)) {
    initStack(context, cdr(arguments));
    push(car(arguments));
  }
}

#define call(TYPE)											\
static TYPE call_##TYPE(TYPE (*f)(), long *v, int n)							\
{													\
  switch (n) {												\
  case  0: return f();											\
  case  1: return f(v[0]);										\
  case  2: return f(v[0],v[1]);										\
  case  3: return f(v[0],v[1],v[2]);									\
  case  4: return f(v[0],v[1],v[2],v[3]);								\
  case  5: return f(v[0],v[1],v[2],v[3],v[4]);								\
  case  6: return f(v[0],v[1],v[2],v[3],v[4],v[5]);							\
  case  7: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6]);						\
  case  8: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7]);						\
  case  9: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8]);					\
  case 10: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9]);					\
  case 11: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10]);				\
  case 12: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11]);			\
  case 13: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12]);		\
  case 14: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12],v[13]);		\
  case 15: return f(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],v[11],v[12],v[13],v[14]);	\
  }													\
  return (TYPE)0;											\
}

typedef void *ptr;

call(ptr)
call(int)
call(long)
call(float)
call(double)

#undef call

static oop call_Pointer(oop func, oop *ap, int nArgs, oop env)
{
  long   v[17];	/* fixme: count the args, then alloca() */
  oop    t= get(func, Pointer,type);
  exec_t f= (exec_t)get(func, Pointer,address);
  int    i= 0;
  oop    targ;
  if (!isPair(t) || (getHead(t) != s_function))		error("cannot call external symbol '%s'", tostring(get(func, Pointer,name)));
  targ= getTail(t);
  t= car(targ);
  while (nArgs-- && i < 15) {
    targ= cdr(targ);
    objectToPrimitive(*ap, v, &i, (targ == nil) ? targ : car(targ));
    ap -= STACK_DIRECTION;
  }
  if (nArgs >= 0)					error("too many arguments in foreign function call");
  if (isPair(t)) {
    if (getHead(t) == s_pointer) {
      oop e= car(getTail(t));
      if (e == s_char)					{ return mkString (call_ptr((void * (*)())f, &v[0], i)        ); }
      else						{ return mkPointer(call_ptr((void * (*)())f, &v[0], i), nil, t); }
    }
  }
  if (!isSymbol(t))					error("cannot receive value from external function '%s'", tostring(get(func, Pointer,name)));
  if (t == s_long)					{ return mkLong (call_long  ((long   (*)())f, &v[0], i)); }
  if (t == s_int || t == s_short || t == s_char)	{ return mkLong (call_int   ((int    (*)())f, &v[0], i)); }
  if (t == s_void)					{ 	         call_int   ((int    (*)())f, &v[0], i) ;  return nil; }
  if (t == s_float)					{ return mkFloat(call_float ((float  (*)())f, &v[0], i)); }
  if (t == s_double)					{ return mkFloat(call_double((double (*)())f, &v[0], i)); }
  error("cannot call external function returning '%s'", tostring(t));
  return nil;
}

#define i(insn) , x_##insn
typedef enum { illegal= 0
insnsDo(i)
} insn_t;
#undef i

static oop compileFunction(oop function);

char *insnName(int opc)
{
  static char *names[]= {
    "illegal" ,
#define i(insn) #insn ,
insnsDo(i)
#undef i
    0
  };
  if (opc > get(s_neq, Symbol,opcode)) opc= 0;
  return names[opc];
}

#if !defined(NDEBUG)
# define i(insn) static long insn##_count= 0;
  insnsDo(i);
# undef i
# define exec_count(insn) ++insn##_count
#else
# define exec_count(insn)
#endif

static oop exec(oop program, oop arguments, oop env)
{
#ifndef NDEBUG
  if (optTrace & 2) { printf("\nEXEC ");  println(program); }
#endif
							GC_PROTECT(env);
  oop  tmp=      nil;					GC_PROTECT(tmp);
  oop  context=  mkContext(nil, env, nil);		GC_PROTECT(context);
  initStack(context, arguments);			assert(isPair(program));
  int  nArgs=    stackSize();
  int  mArgs=    nArgs;
  oop *ap=     	 get(context, Context,sp);
  context=       mkContext(nil, env, context);
  set(context, Context,pc, program);
  oop  code=     getHead(program);
  oop  literals= getTail(program);
  int  ip=       0;
  oop *kelems=   0;

#define loadKernel(C)	kelems= get(get(C, Context,kernel), Array,elements)

#if (CACHE_SP)
  oop *sp;
  oop *sb;
# define loadSP(C)	sp= get(C, Context,sp);  sb= get(C, Context,stackBase);  loadKernel(C)
# define saveSP(C)	    set(C, Context,sp, sp)
# define flushSP(C)	    set(C, Context,sp, sp)
# undef  push
# define push(O)	if (sp == sb) { sp= growStack(context);  sb= get(context, Context,stackBase); } *(sp += STACK_DIRECTION)= (O)
# undef  nos
# define nos(N)		(sp[(N)])
# undef  drop
# define drop(N)	(sp += (N))
# undef  stackSize
# define stackSize()	(get(context, Context,stackLimit) - sp)
  loadSP(context);
#else
# define loadSP(C)	loadKernel(C)
# define saveSP(C)
# define flushSP(C)
#endif

# define parameter	(get(code, WordArray,elements)[ip])
# define operand	(get(literals, Array,elements)[parameter])
# define next		++ip;  continue
# define nextArg(V)	if (nArgs) { V= *ap; ap -= STACK_DIRECTION;  --nArgs; } else { V= nil; }
# define restArgs(V)	if (nArgs) { V= makeArgs(ap, nArgs);  nArgs= 0; } else { V= nil; }
  for (;;) {
    int opc= get(code, WordArray,elements)[ip++];
#  ifndef NDEBUG
    //    if (optTrace & 8) {
    //      printf("\t\t"); println(stack);
    //    }
    if (optTrace & 4) {
      printf("* %04i %02x %02x %s ", ip-1, opc, (int)get(code, WordArray,elements)[ip], insnName(opc));
      if (notNil(operand)) print(operand);
      printf("\n");
    }
#  endif
    switch (opc) {
#   define insn(INSN) case x_##INSN: exec_count(INSN);
    insn(push) {
      push(operand);
      next;
    }
    insn(lookup) {
      oop assoc= assq(operand, env);		if (isNil(assoc)) error("undefined: %s", tostring(operand));
      push(getTail(assoc));
      next;
    }
    insn(assign) {
      oop assoc= assq(operand, env);		if (isNil(assoc)) error("undefined: %s", tostring(operand));
      set(assoc, Pair,tail, tos());
      next;
    }
    insn(dup) {
      oop top= tos();
      push(top);
      next;
    }
    insn(pop) {
      pop();
      next;
    }
    insn(drop) {
      int n= getLong(operand);
      drop(n);
      next;
    }
    insn(bind) {
      flushSP(context);
      tmp= cons(operand, tos());
      pop();
      env= cons(tmp, env);
      next;
    }
    insn(arg) {
      flushSP(context);
      nextArg(tmp);
      tmp= cons(operand, tmp);
      env= cons(tmp, env);
      next;
    }
    insn(rest) {
      flushSP(context);
      restArgs(tmp);
      tmp= cons(operand, tmp);
      env= cons(tmp, env);
      next;
    }
    insn(popenv) {
      int n= getLong(operand);
      while (n--) env= getTail(env);
      next;
    }
    insn(load) {
      oop assoc= operand;
      push(getTail(assoc));
      next;
    }
    insn(name) {
      oop top= tos();
      if (isExpr(top) && isNil(get(top, Expr,name))) {
	set(top, Expr,name, operand);
      }
      next;
    }
    insn(store) {
      oop top= tos();
      oop assoc= operand;
      rplacd(assoc, top);
      next;
    }
    insn(bf) {
      oop top= tos();
      pop();
      if (isNil(top)) { ip= parameter;  continue; }
      next;
    }
    insn(bt) {
      oop top= tos();
      pop();
      if (notNil(top)) { ip= parameter;  continue; }
      next;
    }
    insn(br) {
      ip= parameter;
      continue;
    }
    insn(lambda) {
      flushSP(context);
      tmp= operand;
      if (isNil(get(tmp, Expr,method)) && isNil(compileFunction(tmp))) error("could not compile closure function in lambda");
      tmp= mkExpr(get(tmp, Expr,expr), get(tmp, Expr,params), get(tmp, Expr,body), get(tmp, Expr,method), env);
      push(tmp);
      next;
    }
    insn(tick) {
      //println(get(context, Context,kernel));
      int budget= getLong(kelems[CONTEXT_KERNEL_BUDGET]);
      if (budget) {
	ip += 2;	/* ignore following pop */
	kelems[CONTEXT_KERNEL_BUDGET]= mkLong(budget - 2);
	next;
      }
      /* give control back to the kernel: (kernel[CTX_KERN_ENTRY] kernel) */
      {
	oop kernel= get(context, Context,kernel);
	push(kernel);
	set(context, Context,kernel, kelems[CONTEXT_KERNEL_PARENT]);	/* this MUST BE UNDONE in the user-code kernel entry point function */
	push(kelems[CONTEXT_KERNEL_ENTRY]);
	push(mkLong(1));
      }
      goto xapply;
    }

    xapply:
    insn(xapply) {
      saveSP(context);
      nArgs=    getLong(tos());
      ap=       &nos(2);
      oop func= nos(1);
      switch (getTypeIndex(func)) {
      case iPointer: {
	oop save= threadSpecific;
	threadSpecific= get(context, Context,kernel);
	tmp= call_Pointer(func, ap, nArgs, env);
	threadSpecific= save;
	push(tmp);
	break;
      }
      case iExpr: {
	++ip;		/* save pc of next insn */
	program=  get(func, Expr,method);
	if (isNil(program) && isNil(program= compileFunction(func))) {
	  //warning("could not compile function in apply");
	  tmp= makeArgs(ap, nArgs);
	  tmp= apply(func, tmp, env);
	  push(tmp);
	  continue;
	}
	assert(isPair(program));
	set(context, Context,ip,  ip);
	set(context, Context,env, env);
	code=     getHead(program);	assert(isWordArray(code));
	literals= getTail(program);	assert(isArray(literals));
	ip= 0;
	context= mkContext(func, env, context);
	loadSP(context);
	set(context, Context,pc, program);
	env= get(func, Expr,env);
	backtrace= context;
	continue;
      }
      case iSubr: {
	oop save= threadSpecific;
	set(func, Subr,count, get(func, Subr,count) + 1);
	set(context, Context,ip,   ip);
	set(context, Context,env,  env);
	tmp= makeArgs(ap, nArgs);
	threadSpecific= get(context, Context,kernel);
	tmp= get(func, Subr,exec)(tmp, env);
	threadSpecific= save;
	push(tmp);
	break;
      }
      default:
	printsp(func);
	error("is not applicable");
	break;
      }
      next;
    }
    insn(return)
      /* if (1 != stackSize()) printf("RETURN STACK SIZE %ld\n", stackSize()); */
      assert(isNil(get(get(context, Context,caller), Context,pc)) || 1 == stackSize());
    insn(xreturn) {
      oop result= tos();
      flushSP(context);
      context= get(context, Context,caller);
      loadSP(context);
      backtrace= context;
      program= get(context, Context,pc);
      if (isNil(program)) {
	GC_UNPROTECT(env);
	if (mArgs != stackSize()) {
	  printf("\n");
	  dump("", program);
	  println(arguments);
	  //println(stack);
	  error("stack contained %ld items at end of exec (expected %ld)", stackSize(), mArgs);
	}
	return result;
      }
      assert(isPair(program));
      code=     getHead(program);
      literals= getTail(program);
      env=  	get(context, Context,env);
      ip=   	get(context, Context,ip);
      push(result);
#    ifndef NDEBUG
      if (optTrace & 2) { printf("\nRETURN ");  println(result); }
#    endif
      continue;
    }
    insn(xdrop) {
      oop result= tos();
      int n= getLong(nos(1));
      drop(n + 3);
      push(result);
      next;
    }
    insn(context) {
      set(get(context, Context,caller), Context,callee, nil);
      push(context);
      next;
    }
    insn(resume) {
      {
	oop cx= tos();
	if (!isContext(cx)) {
	  println(cx);
	  error("resuming in non-context object");
	}
	pop();
	tmp= tos();
	saveSP(context);
	context= cx;
	loadSP(context);
      }
      program= get(context, Context,pc);
      if (!isPair(program)) error("resumed context has illegal pc");
      env= 	get(context, Context,env);
      ip=  	get(context, Context,ip);
      code=     getHead(program);
      literals= getTail(program);
      //printf("RESUME STACK SIZE %ld at ip %i\n", get(stack, Array,size), ip);
      push(tmp);
#    ifndef NDEBUG
      if (optTrace & 2) { printf("\nRESUME ");  println(tmp); }
#    endif
      backtrace= context;
      continue;
    }
    insn(addr) {
      flushSP(context);
      oop assoc= assq(operand, env);		if (isNil(assoc)) error("undefined: %s", tostring(operand));
      oop *addr= isPair(assoc) ? &_get(assoc, Pair.tail) : nil;
      push(mkPointer(addr, getHead(assoc), nil));
      next;
    }
    insn(cons) {
      flushSP(context);
      tmp= cons(tos(), nos(1));
      drop(1);
      tos()= tmp;
      ip += 2;
      next;
    }
    insn(car) {
      tos()= car(tos());
      ip += 2;
      next;
    }
    insn(cdr) {
      tos()= cdr(tos());
      ip += 2;
      next;
    }
    insn(setcar) {
      tmp= tos();
      drop(1);
      if (isPair(tmp)) set(tmp, Pair,head, tos());
      ip += 2;
      next;
    }
    insn(setcdr) {
      tmp= tos();
      drop(1);
      if (isPair(tmp)) set(tmp, Pair,tail, tos());
      ip += 2;
      next;
    }
#  define special(OP, ID)				\
    insn(ID) {						\
      oop lhs= tos();					\
      oop rhs= nos(1);					\
      if (isLong(lhs) && isLong(rhs)) {			\
	drop(2);					\
	push(mkLong(getLong(lhs) OP getLong(rhs)));	\
	ip += 2;					\
	next;						\
      }							\
      push(subr_##ID);					\
      push(mkLong(2));					\
      goto xapply;					\
    }
    special(+, 	add);
    special(-, 	sub);
    special(*, 	mul);
    special(/, 	div);
    special(%, 	mod);
    special(<<, asl);
    special(>>, asr);
    special(&, 	and);
    special(|, 	or);
    special(^, 	xor);
#  undef special
#  define special(OP, ID)				\
    insn(ID) {						\
      oop lhs= tos();					\
      oop rhs= nos(1);					\
      if (isLong(lhs) && isLong(rhs)) {			\
	drop(2);					\
	push(mkBoolean(getLong(lhs) OP getLong(rhs)));	\
	ip += 2;					\
	next;						\
      }							\
      push(subr_##ID);					\
      push(mkLong(2));					\
      goto xapply;					\
    }
    special(<, 	lt);
    special(<=, le);
    special(>=,	ge);
    special(>, 	gt);
#  undef special
#  define special(OP, ID)				\
    insn(ID) {						\
      oop lhs= tos();					\
      oop rhs= nos(1);					\
      drop(2);						\
      push(mkBoolean(lhs OP rhs));			\
      ip += 2;						\
      next;						\
    }
    special(==,	eq);
    special(!=, neq);
#  undef special
    insn(eql) {
      oop lhs= tos();
      oop rhs= nos(1);
      drop(2);
      push(mkBoolean(equal(lhs, rhs)));
      ip += 2;
      next;
    }
    insn(neql) {
      oop lhs= tos();
      oop rhs= nos(1);
      drop(2);
      push(mkBoolean(!equal(lhs, rhs)));
      ip += 2;
      next;
    }
    default:
      fprintf(stderr, "\n%d ", opc);
      error("Illegal interpreter instruction");
    }
  }
#undef next
  return nil;
}
