/* gc.c -- trivial single-threaded stop-world non-moving mark-sweep collector
**
** Copyright (c) 2008 Ian Piumarta
** All Rights Reserved
**
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
**
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
**
** Last edited: 2009-09-22 00:46:07 by piumarta on emilia-2.local
*/

#include "gc.h"

#define GC_ALIGN	sizeof(long)
#define GC_MEMORY	0x7fffffff
#define GC_QUANTUM	50*1024
#if defined(DEBUGGC)
# define ALLOCS_PER_GC	1
#else
# define ALLOCS_PER_GC	32768
#endif

#define VERBOSE		0
#define USE_PREV	0

typedef struct _gcheader
{
  unsigned int		size  : 24	__attribute__((__packed__));
  union {
    int			flags :  3;
    struct {
      unsigned int	used  :  1;
      unsigned int	atom  :  1;
      unsigned int	mark  :  1;
    }					__attribute__((__packed__));
  }					__attribute__((__packed__));
  struct _gcheader *next;
#if USE_PREV
  struct _gcheader *prev;
#endif
  struct _gcfinaliser	*finalisers;
} gcheader;

static inline void *hdr2ptr(gcheader *hdr)	{ return (void *)(hdr + 1); }
static inline gcheader *ptr2hdr(void *ptr)	{ return (gcheader *)ptr - 1; }

typedef struct _gcfinaliser
{
  void			*ptr;
  GC_finaliser_t	 finaliser;
  void			*data;
  struct _gcfinaliser	*next;
} gcfinaliser;

static gcheader  gcbase= {
  0
  ,{ -1 }
  ,&gcbase
#if USE_PREV
  ,&gcbase
#endif
};

static gcheader *gcnext= &gcbase;

static size_t	gcQuantum= GC_QUANTUM;
static int	gcCount=   ALLOCS_PER_GC;
static int	gcAllocs=  ALLOCS_PER_GC;
static size_t	gcMemory=  GC_MEMORY;

static gcfinaliser *finalisable= 0;

GC_API void *GC_malloc(size_t lbs)
{
  gcheader *hdr, *org;
  if ((!--gcAllocs) || (gcMemory < lbs)) {
    //fprintf(stderr, "%i %lu %ld\t", gcAllocs, gcMemory, lbs);
#  if VERBOSE >= 1
    if (gcAllocs > 0) fprintf(stderr, "GC: heap full after %i allocations\n", gcCount - gcAllocs);
#  endif
    gcAllocs= gcCount;
    GC_gcollect();
    //fprintf(stderr, "GC %i %lu %ld\n", gcAllocs, gcMemory, lbs);
    if (gcMemory < lbs) goto full;
  }
  org= hdr= gcnext;
  lbs= (lbs + GC_ALIGN-1) & ~(GC_ALIGN-1);
#if VERBOSE > 1
  fprintf(stderr, "malloc %uld\n", lbs);
#endif
 again:
#if VERBOSE > 3
  {
    gcheader *h= gcnext;
    do { 
      fprintf(stderr, "  %2d %p <- %p -> %p = %x\n", h->flags, h->prev, h, h->next, h->size);
      h= h->next;
    } while (h != gcnext);
  }
#endif
  do {
#  if VERBOSE > 2
    fprintf(stderr, "? %2d %p <- %p -> %p = %x\n", hdr->flags, hdr->prev, hdr, hdr->next, hdr->size);
#  endif
    if ((!hdr->used) && (hdr->size >= lbs))
      {
	void *mem;
	if (hdr->size >= lbs + sizeof(gcheader) + GC_ALIGN)
	  {
	    gcheader *ins= (gcheader *)(hdr2ptr(hdr) + lbs);
	    ins->flags= 0;
#	   if USE_PREV
	    ins->prev= hdr;
#	   endif
	    ins->next= hdr->next;
	    ins->size= hdr->size - lbs - sizeof(gcheader);
#	   if USE_PREV
	    hdr->next->prev= ins;
#	   endif
	    hdr->next= ins;
	    hdr->size= lbs;
	  }
	hdr->used= 1;
	hdr->finalisers= 0;
	gcnext= hdr->next;
#      if VERBOSE > 2
	fprintf(stderr, "MALLOC %p\n", hdr);
#      endif
	mem= hdr2ptr(hdr);
	memset(mem, 0, hdr->size);
	gcMemory -= hdr->size;
	//if (mem == (void *)0x200298) { fprintf(stderr, "ALLOCATING %p\n", mem); abort(); }
	return mem;
      }
    hdr= hdr->next;
  } while (hdr != org);
  {
    size_t incr= gcQuantum;
    size_t req= sizeof(gcheader) + lbs;
    while (incr <= req) incr *= 2;
    //fprintf(stderr, "extending by %ld => %ld @ %d\n", req, incr, (int)(gcCount - gcAllocs));
    hdr= (gcheader *)malloc(incr);
    //fprintf(stderr, "buffer at %x\n", (int)hdr);
    if (hdr != (gcheader *)-1)
      {
#      if USE_PREV
	hdr->flags= 0;
	hdr->next= &gcbase;
	hdr->prev= gcbase.prev;
	hdr->prev->next= hdr;
	gcbase.prev= hdr;
	hdr->size= incr - sizeof(gcheader);
#      else
	hdr->flags= 0;
	hdr->next= gcbase.next;
	gcbase.next= hdr;
	hdr->size= incr - sizeof(gcheader);
#      endif
#if VERBOSE
	fprintf(stderr, "extend by %ld at %p\n", hdr->size, hdr);
#endif
	goto again;
      }
    fprintf(stderr, "GC: sbrk failed\n");
  }
 full:
  fprintf(stderr, "GC: out of memory\n");
  abort();
  return 0;
}

GC_API void *GC_malloc_atomic(size_t lbs)
{
  void *mem= GC_malloc(lbs);
  ptr2hdr(mem)->atom= 1;
  return mem;
}

GC_API void *GC_realloc(void *ptr, size_t lbs)
{
  gcheader *hdr= ptr2hdr(ptr);
  void *mem;
  if (lbs <= hdr->size) return ptr;
  mem= GC_malloc(lbs);
  memcpy(mem, ptr, hdr->size);
  ptr2hdr(mem)->atom= hdr->atom;
  GC_free(ptr);
  return mem;
}

static gcheader *GC_freeHeader(gcheader *hdr)
{
#if VERBOSE > 1
  fprintf(stderr, "FREE %p\n", hdr);
#endif
  hdr->flags= 0;
  gcMemory += hdr->size;
#if USE_PREV
  if ((!hdr->prev->flags) && (hdr2ptr(hdr->prev) + hdr->prev->size == hdr))
    {
#    if VERBOSE > 2
      fprintf(stderr, "COALESCE PREV %p\n", hdr->prev);
#    endif
      hdr->prev->next= hdr->next;
      hdr->next->prev= hdr->prev;
      hdr->prev->size += sizeof(gcheader) + hdr->size;
      hdr= hdr->prev;
    }
#endif
  if ((!hdr->next->used) && (hdr2ptr(hdr) + hdr->size == hdr->next))
    {
#    if VERBOSE > 2
      fprintf(stderr, "COALESCE NEXT %p\n", hdr->next);
#    endif
      hdr->size += sizeof(gcheader) + hdr->next->size;
      hdr->next= hdr->next->next;
#    if USE_PREV
      hdr->next->prev= hdr;
#    endif
    }
#if VERBOSE > 3
  {
    gcheader *h= &gcbase;
    do { 
      fprintf(stderr, "  %2d %p <- %p -> %p = %x\n", h->flags, h->prev, h, h->next, h->size);
      h= h->next;
    } while (h != &gcbase);
  }
#endif
  return hdr;
}

GC_API void GC_free(void *ptr)
{
  gcnext= GC_freeHeader(ptr2hdr(ptr));
}

GC_API size_t GC_size(void *ptr)
{
  return ptr2hdr(ptr)->size;
}

GC_API void GC_default_pre_mark_function(void) {}

GC_pre_mark_function_t GC_pre_mark_function= GC_default_pre_mark_function;

GC_API void GC_default_mark_function(void *ptr)
{
  gcheader *hdr= ptr2hdr(ptr);
  void	  **pos= ptr;
  void	  **lim= hdr2ptr(hdr) + hdr->size - sizeof(void *);
  while (pos <= lim)
    {
      void *field= *pos;
      if (field && !((long)field & 1))
	GC_mark(field);
      ++pos;
    }
}

GC_mark_function_t GC_mark_function= GC_default_mark_function;

GC_API void GC_mark(void *ptr)
{
  if ((long)ptr & 1) return;
  gcheader *hdr= ptr2hdr(ptr);
#if VERBOSE > 2
  fprintf(stderr, "mark? %p %d\n", hdr, hdr->flags);
#endif
  if (!hdr->mark) {
    hdr->mark= 1;
    if (!hdr->atom)
      GC_mark_function(ptr);
  }
}

GC_API void GC_mark_leaf(void *ptr)
{
  ptr2hdr(ptr)->mark= 1;
}

GC_free_function_t GC_free_function= 0;

GC_API void GC_sweep(void)
{
  gcheader *hdr= gcbase.next;
#if !USE_PREV
  gcheader *prev= hdr;
#endif
  do {
#if VERBOSE > 2
    fprintf(stderr, "sweep? %p %d\n", hdr, hdr->flags);
#endif
    if (hdr->flags)
      {
	if (hdr->mark)
	  hdr->mark= 0;
	else {
	  if (hdr->finalisers) {
	    while (hdr->finalisers) {
	      gcfinaliser *gcf= hdr->finalisers;
	      hdr->finalisers= gcf->next;
	      gcf->next= finalisable;
	      finalisable= gcf;
	    }
	  }
	  else {
	    if (GC_free_function) GC_free_function(hdr2ptr(hdr));
	    hdr= GC_freeHeader(hdr);
#          if !USE_PREV
	    if ((!prev->flags) && (hdr2ptr(prev) + prev->size == hdr))
	      {
#              if VERBOSE > 2
		fprintf(stderr, "COALESCE PREV %p\n", prev);
#              endif
		prev->next= hdr->next;
		prev->size += sizeof(gcheader) + hdr->size;
		hdr= prev;
	      }
#          endif
	  }
	}
      }
#  if !USE_PREV
    prev= hdr;
#  endif
    hdr= hdr->next;
  } while (hdr != &gcbase);
  gcnext= gcbase.next;
  while (finalisable)
    {
      gcfinaliser *gcf= finalisable;
      gcf->finaliser(gcf->ptr, gcf->data);
      finalisable= gcf->next;
      free(gcf);
    }
}

static void ***roots= 0;
static size_t numRoots= 0;
static size_t maxRoots= 0;

struct GC_StackRoot *GC_stack_roots= 0;

GC_API void GC_add_root(void *root)
{
  if (numRoots == maxRoots)
    roots= maxRoots
      ? realloc(roots, sizeof(roots[0]) * (maxRoots *= 2))
      : malloc (       sizeof(roots[0]) * (maxRoots= 128));
  roots[numRoots++]= (void **)root;
  assert(root);
}

GC_API void GC_delete_root(void *root)
{
  int i;
  for (i= 0;  i < numRoots;  ++i)
    if (roots[i] == (void **)root)
      break;
  if (i < numRoots)
    {
      memmove(roots + i, roots + i + 1, sizeof(roots[0]) * (numRoots - i));
      --numRoots;
    }
}

GC_API long GC_collections= 0;

/*GC_API*/ void GC_gcollect(void)
{
  int i;
  struct GC_StackRoot *sr;
  ++GC_collections;
#if defined(DEBUGGC)
  {
#  undef static
    static char *cursors= "-\\|/";
    static int cursor= 0;
    if (GC_collections % 1000 == 0) {
      if (0 == cursors[cursor]) cursor= 0;
      fprintf(stderr, "%c\010", cursors[cursor]);
      ++cursor;
    }
#  if (NONSTATIC)
#    define static
#  endif
  }
#endif
  GC_pre_mark_function();
#if VERBOSE >= 1
  fprintf(stderr, "*** GC: mark roots\n");
#endif
  for (i= 0;  i < numRoots;  ++i)		if (*roots[i])   GC_mark(*roots[i]);
#if VERBOSE > 0
  fprintf(stderr, "*** GC: mark stack\n");
#endif
  for (sr= GC_stack_roots;  sr;  sr= sr->next)	{
    if (*(sr->root)) GC_mark(*(sr->root));
  }
#if VERBOSE > 0
  fprintf(stderr, "*** GC: sweep\n");
#endif
  GC_sweep();
#if VERBOSE > 0
  fprintf(stderr, "*** GC: done\n");
#endif
}

/*GC_API*/ size_t GC_count_objects(void)
{
  gcheader *hdr= gcbase.next;
  size_t count= 0;
  do {
    if (hdr->flags)
      ++count;
    hdr= hdr->next;
  } while (hdr != &gcbase);
  return count;
}

/*GC_API*/ size_t GC_count_bytes(void)
{
  gcheader *hdr= gcbase.next;
  size_t count= 0;
  do {
    if (hdr->flags)
      count += hdr->size;
    hdr= hdr->next;
  } while (hdr != &gcbase);
  return count;
}

GC_API int GC_check(void *ptr)
{
  gcheader *hdr= ptr2hdr(ptr);
  if (!hdr->used) {
    hdr->used= 1;
    return 0;
  }
  return 1;
}

/*GC_API*/ void GC_register_finaliser(void *ptr, GC_finaliser_t finaliser, void *data)
{
  gcheader    *gch = ptr2hdr(ptr);
  gcfinaliser *gcf = (struct _gcfinaliser *)malloc(sizeof(struct _gcfinaliser));
  gcf->ptr         = ptr;
  gcf->finaliser   = finaliser;
  gcf->data        = data;
  gcf->next        = gch->finalisers;
  gch->finalisers  = gcf;
}


#if 0

#include <stdlib.h>

#define RAND(N)	(1 + (int)((float)N * (rand() / (RAND_MAX + 1.0))))

struct cell { int tag;  struct cell *next; };

void *mklist(int n)
{
  struct cell *cell;
  if (!n) return 0;
  cell= GC_malloc(8);
  cell->tag= n << 1 | 1;
  cell->next= mklist(n - 1);
  return cell;
}

void delist(struct cell *cell)
{
  if (cell && cell->next && cell->next->next) {
    cell->next= cell->next->next;
    delist(cell->next->next);
  }
}

int main()
{
  int i, j;
  void *a, *b, *c, *d, *e;
  for (i= 0;  i < 1000000;  ++i) {
#if !VERBOSE
# define printf(...)
#endif
    //#define GC_malloc malloc
    //#define GC_free free
    a= GC_malloc(RAND(1));	    printf("%p\n", a);
    b= GC_malloc(RAND(10));	    printf("%p\n", b);
    c= GC_malloc(RAND(100));	    printf("%p\n", c);
    d= GC_malloc(RAND(1000));	    printf("%p\n", d);
    e= GC_malloc(RAND(10000));	    printf("%p\n", e);
    GC_free(a);
    GC_free(b);
    //    GC_free(c);
    GC_free(d);
    GC_free(e);
    a= GC_malloc(RAND(100));	    printf("%p\n", a);
    b= GC_malloc(RAND(200));	    printf("%p\n", b);
    c= GC_malloc(RAND(300));	    printf("%p\n", c);
    d= GC_malloc(RAND(400));	    printf("%p\n", d);
    e= GC_malloc(RAND(500));	    printf("%p\n", e);
    GC_free(e);
    GC_free(d);
    //    GC_free(c);
    GC_free(b);
    GC_free(a);
    a= GC_malloc(RAND(4));	    printf("%p\n", a);
    b= GC_malloc(RAND(16));	    printf("%p\n", b);
    c= GC_malloc(RAND(64));	    printf("%p\n", c);
    d= GC_malloc(RAND(256));	    printf("%p\n", d);
    e= GC_malloc(RAND(1024));	    printf("%p\n", e);
    GC_free(e);
    GC_free(b);
    //    GC_free(c);
    GC_free(d);
    GC_free(a);
    a= GC_malloc(RAND(713));	    printf("%p\n", a);
    b= GC_malloc(RAND(713));	    printf("%p\n", b);
    c= GC_malloc(RAND(713));	    printf("%p\n", c);
    d= GC_malloc(RAND(713));	    printf("%p\n", d);
    e= GC_malloc(RAND(713));	    printf("%p\n", e);
    GC_free(a);
    GC_free(c);
    //    GC_free(e);
    GC_free(d);
    GC_free(b);
#undef printf
    if (i % 100000 == 0) printf("alive: %d bytes in %d objects\n", GC_count_bytes(), GC_count_objects());
    GC_sweep();
    if (i % 100000 == 0) printf("alive: %d bytes in %d objects\n", GC_count_bytes(), GC_count_objects());
  }
  {
    GC_PROTECT(a);
    for (i= 0;  i < 100;  ++i) {
      for (j= 0;  j < 100;  ++j) {
	a= mklist(2000);
	delist(a);
#if VERBOSE
	{
	  struct cell *c= a;
	  printf("----\n");
	  while (c) {
	    printf("%p %d %p\n", c, c->tag >> 1, c->next);
	    c= c->next;
	  }
	}
#endif
      }
      GC_gcollect();
    }
    GC_UNPROTECT(a);
  }
  printf("alive: %d bytes in %d objects\n", GC_count_bytes(), GC_count_objects());
  GC_sweep();
  printf("alive: %d bytes in %d objects\n", GC_count_bytes(), GC_count_objects());
  return 0;
}

#endif
