/* intrinsic.c -- s-expression primitives for the implementation to use
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:35 by piumarta on emilia-2.local
*/

static char *tostring(oop obj);

static int eq(oop a, oop b)
{
  return a == b;
}

static int equal(oop a, oop b)
{
  if (eq(a, b)) return 1;
  if (getTypeIndex(a) != getTypeIndex(b)) return 0;
  switch (getTypeIndex(a)) {
  case iUndefined:	break;
  case iTag:		return eq(a, b);
  case iLong:		return getLong(a) == getLong(b);
  case iFloat:		return getFloat(a) == getFloat(b);
  case iPointer:	return get(a, Pointer,address) == get(b, Pointer,address);
  case iPair:		return equal(getHead(a), getHead(b)) && equal(getTail(a), getTail(b));
  case iString:		return (get(a, String,size) == get(b, String,size))
			    && !memcmp(get(a, String,elements), get(b, String,elements), getLong(get(a, String,size)));
  case iSymbol:		break;
  case iArray: {
    size_t sa= getLong(get(a, Array,size)),      sb= getLong(get(b, Array,size));
    if (sa != sb) return 0;
    oop   *ea= get(a, Array,elements),		*eb= get(b, Array,elements);
    size_t i;
    for (i= 0;  i < sa;  ++i) if (!equal(ea[i], eb[i])) return 0;
    return 1;
  }
  case iByteArray: {
    size_t sa= getLong(get(a, ByteArray,size)),  sb= getLong(get(b, ByteArray,size));
    if (sa != sb) return 0;
    byte *ea= get(a, ByteArray,elements),	*eb= get(b, ByteArray,elements);
    size_t i, j;
    for (i= 0, j= 0;  i < sa;  ++i, ++j) if (ea[i] != eb[j]) return 0;
    return 1;
  }
  case iWordArray: {
    size_t sa= getLong(get(a, WordArray,size)),  sb= getLong(get(b, WordArray,size));
    if (sa != sb) return 0;
    word *ea= get(a, WordArray,elements),	*eb= get(b, WordArray,elements);
    size_t i, j;
    for (i= 0, j= 0;  ; ++i, ++j) if (ea[i] != eb[j]) return 0;
    return 1;
  }
  case iSyntax:		return equal(get(a, Syntax,compiler), get(b, Syntax,compiler));
  case iExpr:		return equal(get(a, Expr,params), get(b, Expr,params))
			    && equal(get(a, Expr,body),   get(b, Expr,body))
			    && eq   (get(a, Expr,env),    get(b, Expr,env));
  case iSubr:		return get(a, Subr,exec) == get(b, Subr,exec);
  case iType:		return eq(a, b);
  case iContext:	return eq(a, b);
  }
  return eq(a, b);
}

static oop assoc(oop key, oop list)
{
  for (;  isPair(list);  list= getTail(list))
    if (equal(key, car(getHead(list))))
      return getHead(list);
  return nil;
}

static oop assq(oop key, oop list)
{
  for (;  isPair(list);  list= getTail(list))
    if (eq(key, car(getHead(list))))
      return getHead(list);
  return nil;
}

static oop nssq(oop key, oop list)
{
  for (;  isPair(list) && isPair(getHead(list));  list= getTail(list))
    if (eq(key, getHead(getHead(list))))
      return getHead(list);
  list= cdr(list);
  for (;  isPair(list) && isPair(getHead(list));  list= getTail(list))
    if (eq(key, getHead(getHead(list))))
      return getHead(list);
  return nil;
}

//static oop memq(oop item, oop list)
//{
//  for (;  isPair(list);  list= getTail(list))
//    if (eq(item, getHead(list)))
//      break;
//  return list;
//}

#if defined(DEBUG)
# define newSize(SIZE) ((SIZE) + 1)
#else
# define newSize(SIZE) ((SIZE) ? ((SIZE) * 2) : 4)
#endif

/* static int string_append(oop string, int obj) */
/* { */
/*   if (!isString(string)) error("non-string argument in string-append"); */
/*   char *elements= get(string, String,elements); */
/*   long sz= get(string, String,size); */
/*   assert(sz + 1 <= get(string, String,capacity)); */
/*   if (sz + 1 == get(string, String,capacity)) { */
/*     GC_PROTECT(string); */
/*     elements= set(string, String,elements, GC_realloc(elements, sizeof(byte) * newSize(sz))); */
/*     set(string, String,capacity, newSize(sz)); */
/*     GC_UNPROTECT(string); */
/*   } */
/*   set(string, String,size, sz + 1); */
/*   elements[sz]= obj; */
/*   elements[sz + 1]= 0; */
/*   return obj; */
/* } */

static oop array_append(oop array, oop obj)
{
  if (!isArray(array)) return error("non-array in append");
  oop *elements= get(array, Array,elements);
  long sz= getLong(get(array, Array,size));
  if (sz == GC_size(elements) / sizeof(oop)) {
    GC_PROTECT(array);
    GC_PROTECT(obj);
    elements= set(array, Array,elements, GC_realloc(elements, sizeof(oop) * newSize(sz)));
    GC_UNPROTECT(array);
  }
  set(array, Array,size, mkLong(sz + 1));
  return elements[sz]= obj;
}

/* static int byteArray_append(oop array, int obj) */
/* { */
/*   if (!isByteArray(array))	error("non-byte-array in append-byte"); */
/*   if (obj < 0 || obj > 255)	error("argument out of range append-byte"); */
/*   byte *elements= get(array, ByteArray,elements); */
/*   long  sz= get(array, ByteArray,size); */
/*   if (sz == get(array, ByteArray,capacity)) { */
/*     GC_PROTECT(array); */
/*     elements= set(array, ByteArray,elements, GC_realloc(elements, newSize(sz))); */
/*     set(array, ByteArray,capacity, newSize(sz)); */
/*     GC_UNPROTECT(array); */
/*   } */
/*   set(array, ByteArray,size, sz + 1); */
/*   return elements[sz]= obj; */
/* } */

/* static int wordArray_append(oop array, int obj) */
/* { */
/*   if (!isWordArray(array))	error("non-word-array in append-word"); */
/*   if (obj < 0 || obj > 255)	error("argument out of range append-word"); */
/*   word *elements= get(array, WordArray,elements); */
/*   long  sz= get(array, WordArray,size); */
/*   if (sz == get(array, WordArray,capacity)) { */
/*     GC_PROTECT(array); */
/*     elements= set(array, WordArray,elements, GC_realloc(elements, sizeof(word) * newSize(sz))); */
/*     set(array, WordArray,capacity, newSize(sz)); */
/*     GC_UNPROTECT(array); */
/*   } */
/*   set(array, WordArray,size, sz + 1); */
/*   return elements[sz]= obj; */
/* } */

#undef newSize

static char *tostring(oop obj)
{
  switch (getTypeIndex(obj)) {
  case iString:	return get(obj, String,elements);
  case iSymbol:	return tostring(get(obj, Symbol,string));
  default:
    fprintf(stderr, "\nNO STRING VALUE FOR ");
    fprintln(obj, stderr);
    return "?";
  }
}

static int list_length(oop list)
{
  int n;
  for (n= 0;  isPair(list);  list= getTail(list)) ++n;
  return n;
}

/* static oop list_append(oop list, oop appendage) */
/* { */
/*   if (isNil(list)) return appendage; */
/*   oop ptr; */
/*   for (ptr= list;  isPair(getTail(ptr));  ptr= getTail(ptr)); */
/*   rplacd(ptr, appendage); */
/*   return list; */
/* } */

static oop list_concat(oop list, oop appendage)
{
  if (!isPair(list)) return isNil(appendage) ? list : appendage;
  oop tail= list_concat(cdr(list), appendage);	GC_PROTECT(tail);
  tail= cons(car(list), tail);			GC_UNPROTECT(tail);
  return tail;
}

static oop makeNamespace(oop env, oop name)
{
  GC_PROTECT(env);
  env= mkPair(nil, env);
  if (notNil(name)) {
    oop assoc= mkPair(name, env);
    GC_PROTECT(assoc);
    assoc= mkPair(assoc, getTail(env));
    rplacd(env, assoc);
    GC_UNPROTECT(assoc);
  }
  GC_UNPROTECT(env);
  return env;
}

static oop localNamespace(oop env)
{
  //fprintf(stderr, "localNamespace from ");  fprintln(env, stderr);
  while (isPair(env) && isPair(getHead(env))) {
    //fprintf(stderr, "  skip ");  fprintln(getHead(env), stderr);
    env= getTail(env);
  }
  //fprintf(stderr, "  answer ");  fprintln(env, stderr);
  return env;
}

static oop declare(oop name, oop env)
{
  oop ns= localNamespace(env);
  oop assoc= nssq(name, ns);
  if (isNil(assoc)) {
    assert(isPair(ns));
    GC_PROTECT(assoc);
    assoc= mkPair(name, nil);
    rplacd(ns, mkPair(assoc, cdr(ns)));
    GC_UNPROTECT(assoc);
  }
  return assoc;
}

static oop define(oop name, oop value, oop env)
{
  return rplacd(declare(name, env), value);
}
