/* reader.c -- convert a stream of characters into s-expressions
** 
** Copyright (c) 2009 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
** 
** Last edited: 2010-08-03 11:14:29 by piumarta on emilia-2.local
*/

typedef oop (*reader)(INPUT *, int, oop);

static oop readIllegal(INPUT *s, int c, oop env);
static oop readBlank(INPUT *s, int c, oop env);
static oop readSymbol(INPUT *s, int c, oop env);
static oop readString(INPUT *s, int c, oop env);
static oop readQuote(INPUT *s, int c, oop env);
static oop readLiteral(INPUT *s, int c, oop env);
static oop readList(INPUT *s, int c, oop env);
static oop readClose(INPUT *s, int c, oop env);
static oop readNumber(INPUT *s, int c, oop env);
static oop readSign(INPUT *s, int c, oop env);
static oop readComment(INPUT *s, int c, oop env);
static oop readObject(INPUT *s, oop env);

#define X readIllegal
#define C readComment
#define B readBlank
#define R readClose
#define N readNumber
#define M readSign
#define S readSymbol
#define T readString
#define L readLiteral
#define P readList
#define Q readQuote

static reader readers[128]= {
   X, /*nul*/  X, /*soh*/  X, /*stx*/  X, /*etx*/  X, /*eot*/  X, /*enq*/  X, /*ack*/  X, /*bel*/
   X, /*bs */  B, /*ht */  B, /*nl */  X, /*vt */  X, /*np */  B, /*cr */  X, /*so */  X, /*si */
   X, /*dle*/  X, /*dc1*/  X, /*dc2*/  X, /*dc3*/  X, /*dc4*/  X, /*nak*/  X, /*syn*/  X, /*etb*/
   X, /*can*/  X, /*em */  X, /*sub*/  X, /*esc*/  X, /*fs */  X, /*gs */  X, /*rs */  X, /*us */
   B, /*sp */  S, /* ! */  T, /* " */  L, /* # */  S, /* $ */  S, /* % */  S, /* & */  Q, /* ' */
   P, /* ( */  R, /* ) */  S, /* * */  M, /* + */  Q, /* , */  M, /* - */  S, /* . */  S, /* / */
   N, /* 0 */  N, /* 1 */  N, /* 2 */  N, /* 3 */  N, /* 4 */  N, /* 5 */  N, /* 6 */  N, /* 7 */
   N, /* 8 */  N, /* 9 */  S, /* : */  C, /* ; */  S, /* < */  S, /* = */  S, /* > */  S, /* ? */
   S, /* @ */  S, /* A */  S, /* B */  S, /* C */  S, /* D */  S, /* E */  S, /* F */  S, /* G */
   S, /* H */  S, /* I */  S, /* J */  S, /* K */  S, /* L */  S, /* M */  S, /* N */  S, /* O */
   S, /* P */  S, /* Q */  S, /* R */  S, /* S */  S, /* T */  S, /* U */  S, /* V */  S, /* W */
   S, /* X */  S, /* Y */  S, /* Z */  X, /* [ */  S, /* \ */  X, /* ] */  S, /* ^ */  S, /* _ */
   Q, /* ` */  S, /* a */  S, /* b */  S, /* c */  S, /* d */  S, /* e */  S, /* f */  S, /* g */
   S, /* h */  S, /* i */  S, /* j */  S, /* k */  S, /* l */  S, /* m */  S, /* n */  S, /* o */
   S, /* p */  S, /* q */  S, /* r */  S, /* s */  S, /* t */  S, /* u */  S, /* v */  S, /* w */
   S, /* x */  S, /* y */  S, /* z */  X, /* { */  S, /* | */  X, /* } */  S, /* ~ */  X, /*del*/
};

#undef X
#undef I
#undef C
#undef B
#undef D
#undef R
#undef N
#undef M
#undef Y
#undef S
#undef L
#undef Q

/* the following look neither like pointers nor like tagged integers */

#define readerEOF	((oop)-2)
#define readerNone	((oop)-6)
#define readerDot	((oop)-10)
#define readerClose	((oop)-14)

static int isprint(int c)
{
  return ' ' <= c && c <= '~';
}

static oop readIllegal(INPUT *s, int c, oop env)
{
  error("illegal character '%c' (%d)\n", isprint(c) ? c : '?', c);
  return readerNone;
}

static oop readInternal(INPUT *s, oop env)
{
  int c;
  while ((c= igetc(s)) >= 0) {
    reader r= (c < (sizeof(readers) / sizeof(*readers))) ? readers[c] : readIllegal;
    oop obj= r(s, c, env);
    if (readerNone != obj) { return obj; }
  }
  return (oop)readerEOF;
}

static oop readObject(INPUT *s, oop env)
{
  oop obj= readInternal(s, env);
  if (readerDot   == obj) { error("unexpected '.' outside list");  return nil; }
  if (readerClose == obj) { error("unexpected ')' outside list");  return nil; }
  return obj;
}

static oop readComment(INPUT *s, int c, oop env)
{
  while ((c= igetc(s)) >= 0 && ('\n' != c && '\r' != c))
    ;
  return readerNone;
}

static oop readBlank(INPUT *s, int c, oop env)	{ return readerNone; }
static oop readClose(INPUT *s, int c, oop env)	{ return readerClose; }

static int digitValue(int c)
{
  switch (c) {
  case '0' ... '9': return c - '0';
  case 'A' ... 'F': return c - 'A' + 10;
  case 'a' ... 'f': return c - 'a' + 10;
  };
  return -1;
}

static oop readLong(INPUT *s, int c, int base, oop env)
{
  long n= c - '0', d;
  while ((c= igetc(s)) > 0 && ((d= digitValue(c)) >= 0) && (d < base))
    n= n * base + d;
  unigetc(c, s);
  return mkLong(n);
}

static oop readNumber(INPUT *s, int c, oop env)
{
  if ('0' == c) {
    int d= igetc(s);
    switch (d) {
    case 'x': case 'X':	return readLong(s, c, 16, env);
    case 'b': case 'B':	return readLong(s, c,  2, env);
    case '0' ... '7':	return readLong(s, d,  8, env);
    }
    unigetc(d, s);
  }
  long n= c - '0';
  while ((c= igetc(s)) > 0 && readers[c] == readNumber) n= n * 10 + c - '0';
  if ('.' == c) {
    double f= 0.0, m= 1.0;
    while ((c= igetc(s)) > 0 && readers[c] == readNumber) f += ((m /= 10.0) * (c - '0'));
    f= (double)n + f;
    if ('e' == c) {
      int es= 0, e= 0;
      c= igetc(s);
      if ('-' == c) {
	c= igetc(s);
	es= 1;
      }
      else if ('+' == c) {
	c= igetc(s);
      }
      while (c > 0 && readers[c] == readNumber) {
	e= e * 10 + c - '0';
	c= igetc(s);
      }
      m= 1.0;
      while (e--) m *=  10.0L;
      f= es ? f / m : f * m;
    }
    unigetc(c, s);
    return mkFloat(f);
  }
  unigetc(c, s);
  return mkLong(n);
}

static oop readSign(INPUT *s, int c, oop env)
{
  int d= igetc(s);
  if (d > 0 && readers[d] == readNumber) {
    oop num= readNumber(s, d, env);
    if ('-' == c) {
      GC_PROTECT(num);
      switch (getTypeIndex(num)) {
      case iLong:   num= mkLong(-getLong(num));		break;
      case iFloat:  num= mkFloat(-getFloat(num));	break;
      default:
	error("non-number read after sign");
	break;
      }
      GC_UNPROTECT(num);
    }
    return num;
  }
  unigetc(d, s);
  return readSymbol(s, c, env);
}

static oop readSymbol(INPUT *s, int c, oop env)
{
  int  size= 1024;
  char *sym= alloca(size + 1);
  int n= 0;
  sym[n++]= c;
  while ((c= igetc(s)) >= 0 && (readers[c] == readSymbol || readers[c] == readNumber || readers[c] == readSign)) {
    if (n == size) {
      char *sym2= alloca(size * 2 + 1);
      memcpy(sym2, sym, size);
      sym= sym2;
      size *= 2;
    }
    sym[n++]= c;
  }
  unigetc(c, s);
  sym[n]= '\0';
  if (!strcmp(sym, ".")) return readerDot;
  return mkSymbol(sym);
}

static int escaped(INPUT *s, int c)
{
  switch (c) {
  case '"':	 		break;
  case '\'':	 		break;
  case '\\':			break;
  case 'a':	 c= '\a';	break;
  case 'b':	 c= '\b';	break;
  case 'e':	 c= '\e';	break;
  case 'f':	 c= '\f';	break;
  case 'n':	 c= '\n';	break;
  case 'r':	 c= '\r';	break;
  case 't':	 c= '\t';	break;
  case 'v':	 c= '\v';	break;
  case 'x': {
    c= digitValue(igetc(s));
    int d= digitValue(igetc(s));
    if (c < 0 || c > 15 || d < 0 || d > 15)
      error("illegal hex digit in escape sequence");
    return c * 16 + d;
  }
  case '0' ... '7': {
    c= digitValue(c);
    int d= digitValue(igetc(s));
    int e= digitValue(igetc(s));
    if (c < 0 || c > 3 || d < 0 || d > 7 || e < 0 || e > 7)
      error("illegal octal digit in escape sequence %d %d %d", c, d, e);
    return c * 64 + d * 8 + e;
  }
  default:
    if (c < 0)
      error("EOF in character escape");
    else
      warning("unknown escape \\%c in literal", c);
    break;
  }
  return c;
}

static oop readString(INPUT *s, int c, oop env)
{
#undef static
  static int capacity= 0;
  static char *str= 0;
#if (NONSTATIC)
# define static
#endif
  int size= 0;
 next:
  while ((c= igetc(s)) >= 0 && (readers[c] != readString)) {
    if ('\\' == c) {
      switch ((c= igetc(s))) {
      case '\n': while ('\r' == (c= igetc(s)));  unigetc(c, s);  goto next;
      case '\r': while ('\n' == (c= igetc(s)));  unigetc(c, s);  goto next;
      default:
	c= escaped(s, c);
	if (c < 0) continue;
	break;
      }
    }
    if (size == capacity) { str= capacity ? realloc(str, capacity *= 2) : malloc(capacity= 64); }
    str[size++]= c;
  }
  if (size == capacity) { str= capacity ? realloc(str, capacity *= 2) : malloc(capacity= 64); }
  str[size]= '\0';
  return mkString(str);
}

static oop readLiteral(INPUT *s, int c, oop env)
{
  switch ((c= igetc(s))) {
  case '!':
    return readComment(s, c, env);
  case '\'':
    switch ((c= igetc(s))) {
    case '\\':
      c= escaped(s, igetc(s));
    default:
      return mkLong(c);
    }
  case '`': {
    oop expr= readObject(s, env);
    if (readerEOF == expr) return error("EOF in #( expression");
    GC_PROTECT(expr);
    expr= eval(expr, env);
    GC_UNPROTECT(expr);
    return expr;
  }
  default:
    break;
  }
  return error("illegal #-expression");
}

static oop readList(INPUT *s, int c, oop env)
{
  oop head= nil, tail, o= nil;
  GC_PROTECT(head);
  GC_PROTECT(o);

  o= readInternal(s, env);
  if (readerClose == o) goto done;
  if (readerEOF   == o) goto errorEOF;
  if (readerDot   == o) goto errorDot;
  head= tail= mkPair(o, nil);
  for (;;) {
    o= readInternal(s, env);
    if (readerClose == o) goto done;
    if (readerEOF   == o) goto errorEOF;
    if (readerDot   == o) break;
    tail= rplacd(tail, mkPair(o, nil));
  }
  o= nil;			/* don't mark the sentinel */
  o= readInternal(s, env);
  if (readerClose == o) goto errorClose;
  if (readerEOF   == o) goto errorEOF;
  if (readerDot   == o) goto errorDot;
  rplacd(tail, o);
  o= readInternal(s, env);
  if (readerClose == o) goto done;
  if (readerEOF   == o) goto errorEOF;
  if (readerDot   == o) goto errorDot;
	     error("missing ')' at end of list");	  head= nil;  goto done;
 errorEOF:   error("EOF while reading list");		  head= nil;  goto done;
 errorDot:   error("unexpected '.' while reading list");  head= nil;  goto done;
 errorClose: error("unexpected ')' while reading list");  head= nil;  goto done;
 done:
  GC_UNPROTECT(head);
  return head;
}

static oop readQuote(INPUT *s, int c, oop env)
{
  oop sym= nil;
  switch (c) {
  case '\'':
    sym= s_quote;
    break;
  case  '`':
    sym= s_quasiquote;
    break;
  case  ',':
    sym= s_unquote;
    if ('@' == (c= igetc(s)))
      sym= s_unquote_splicing;
    else
      unigetc(c, s);
    break;
  default:
    error("unexpected character in readQuote");
    break;
  }
  oop obj= readObject(s, env);
  if (readerEOF == obj) return sym;
  GC_PROTECT(obj);
  obj= mkPair(obj, nil);
  obj= mkPair(sym, obj);
  GC_UNPROTECT(obj);
  return obj;
}
